/*
  ####################################################################
  REQ: estimating  the  rate of  elementary  quartets  (REQ)  for each 
       branch of a phylogenetic tree from a distance matrix

  Copyright (C) 2017,2018,2019  Alexis Criscuolo 

  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation,  either version 3  of the License,  or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,  but
  WITHOUT  ANY  WARRANTY;   without  even  the   implied  warranty  of
  MERCHANTABILITY or  FITNESS FOR  A PARTICULAR  PURPOSE.  See the GNU 
  General Public License for more details.
  
  You should have  received a copy  of the GNU  General Public License
  along with this program. If not, see <http://www.gnu.org/licenses/>.

  Contact: 
  Institut Pasteur
  Bioinformatics and Biostatistics Hub
  C3BI, USR 3756 IP CNRS
  Paris, FRANCE

  alexis.criscuolo@pasteur.fr
  ####################################################################
*/

import java.io.*;
import java.util.*;
import java.text.*;

public class REQ {

    // constants
    final static String VERSION = "v1.3.190304ac";
    final static int INF = Integer.MAX_VALUE;
    
    // io
    static BufferedReader in;
    static BufferedWriter out;
    static File dfile, tfile;
    static boolean verbose;
    static NumberFormat df;

    // data
    static int n;
    static ArrayList<String> lbl;
    static float[][] dm;
    static StringBuilder nwk, tr, tro;

    // stuffs
    static int i, j, m, u, v, w, x, y, last, sup;
    static double re, ab, ac, bc, abcd, up, dn;
    static short si;
    static String line;
    static String[] split;
    static short[] sta, stb, stc, std, sa;
    static TreeSet<Short> ts, stdts;
    static ArrayList<Double> are;

    public static void main(String[] args) throws IOException {

	//### man ############################################################################################################################################################################
	if ( args.length < 3 ) {
	    System.out.println("");
	    System.out.println(" REQ " + VERSION);
	    System.out.println("");
	    System.out.println(" USAGE:  REQ  <dfile>  <tfile>  <outfile>  [-v]");
	    System.out.println("");
	    System.out.println("   <dfile>    distance matrix file in either PHYLIP lower-triangular or");
	    System.out.println("              square format");
	    System.out.println("   <tfile>    unrooted binary phylogenetic tree file with no confidence");
	    System.out.println("              value at branches in NEWICK format");
	    System.out.println("   <outfile>  outfile name");
	    System.out.println("   -v         verbose mode");
	    System.out.println("");
	    System.exit(0);
	}

	//### init ###########################################################################################################################################################################
        df = NumberFormat.getNumberInstance(Locale.US); df.setGroupingUsed(false); df.setMaximumFractionDigits(3); df.setMinimumFractionDigits(3);
	verbose = ( (args.length > 3) && args[3].equals("-v") ) ? true : false;
	if ( verbose ) System.out.println("REQ " + VERSION);
	if ( ! (dfile = new File(args[0])).exists() ) { System.err.println("distance matrix file does not exist: " + args[0]); System.exit(1); }
	if ( ! (tfile = new File(args[1])).exists() ) { System.err.println("tree file does not exist: " + args[1]); System.exit(1); }
		
	//### reading distance matrix dm #####################################################################################################################################################
	in = new BufferedReader(new FileReader(dfile));
	while ( true ) try { if ( (line=in.readLine().trim()).length() != 0 ) break; } catch ( NullPointerException e ) { System.out.println("matrix file is empty"); System.exit(1); }
	try { n = Integer.parseInt(line); } catch ( NumberFormatException e ) { System.out.println("distance matrix file is incorrectly formatted: " + args[0]); System.exit(1); }
	if ( n > 32760 ) { System.out.println("too many taxa (>32760)"); System.exit(1); }
	lbl = new ArrayList<String>(n); dm = new float[n][]; i = -1;
	while ( true ) {
	    try { line = in.readLine().trim(); } catch ( NullPointerException e ) { in.close(); break; }
	    split = line.split("\\s+"); lbl.add(split[0]); dm[++i] = new float[i]; j = 0; while ( ++j <= i ) dm[i][--j] = Float.parseFloat(split[++j]);
	}
	if ( ++i != n ) { System.out.println("distance matrix file is incorrectly formatted: " + args[0]); System.exit(1); }
	
	//### reading phylogenetic tree nwk ##################################################################################################################################################
	nwk = new StringBuilder(""); in = new BufferedReader(new FileReader(tfile));
	while ( true ) try { nwk = nwk.append(in.readLine().trim()); } catch ( NullPointerException e ) { in.close(); break; }
	tr = new StringBuilder(nwk.toString());
	
	//### discarding branch lengths from tr ##############################################################################################################################################
	//u = -1; while ( (u=tr.indexOf(":", ++u)) != -1 ) { x = ( (x=tr.indexOf(",", u)) != -1 ) ? x : INF; y = ( (y=tr.indexOf(")", u)) != -1 ) ? y : INF; tr = tr.delete(u, (x<y)?x:y); }
	u = -1 ; while ( (u=tr.indexOf(":", ++u)) != -1 ) tr = tr.delete(u, (x=((x=tr.indexOf(",",u))!=-1)?x:INF) < (y=((y=tr.indexOf(")",u))!=-1)?y:INF) ? x : y);
	
	//### encoding lbl within tr #########################################################################################################################################################
	u = -1;
	while ( (u = (x=((x=tr.indexOf(",",u))!=-1)?x:INF) < (y=((y=tr.indexOf("(",u))!=-1)?y:INF) ? x : y) < INF ) {
	    if ( tr.charAt(++u) == '(' ) continue;
	    tr = tr.replace(u, (v = (x=((x=tr.indexOf(",",u))!=-1)?x:INF) < (y=((y=tr.indexOf(")",u))!=-1)?y:INF) ? x : y), String.valueOf(j=lbl.indexOf(tr.substring(u, v))));
	    if ( j == -1 ) { System.out.println(tr.substring(u, v) + " is not inside distance matrix"); System.exit(1); }
	}
	if ( verbose ) System.out.println("# " + tr.toString());
	    
	//### estimating the rate of EQ re for each internal branch e ########################################################################################################################
	tro = new StringBuilder(tr.toString());	
	//# first tr rooting:   ((ST1),(ST2),(ST3));  =>  (((ST1),(ST2)),(ST3));  ###############################################################
	//#                                                            |      |
	//#                                                           sup    last
	tr = tr.insert(0, '('); sup = apc(tr, tr.lastIndexOf(")")); tr = tr.insert(sup, ')'); //# NOTE: closing parenthesis at index 'sup' should not be considered for REQ calculations
	last = tr.lastIndexOf(")"); 
	if ( verbose ) { i = -1; while ( ++i < n ) System.out.println(i + ": " + lbl.get(i)); /*System.out.println("# " + tr.toString());*/ }

	ts = new TreeSet<Short>(); si = 0; --si; while ( ++si < n ) ts.add(new Short(si)); are = new ArrayList<Double>(); u = -1;
	while ( (u=tr.indexOf(")", ++u)) != last ) {
	    //# second tr rooting:   ((ST1),(ST2),(ST3));  =>  ((ST1),((ST2),(ST3)));  ##########################################################
	    //#                                                             |     |
	    //#                                                             u    last
	    if ( u == sup ) {
		last = tro.lastIndexOf(")"); tr = tro.insert(last, ')'); v = apc(tr, apc(tr, last)); tr = tr.insert(++v, '(');
		++last; //# NOTE: closing parenthesis at index 'last' should not be considered for REQ calculations
		/*if ( verbose ) System.out.println("# " + tr.toString());*/
		sup = 0; --u; continue;
	    }
	    //# parsing every internal branch e at index u in order to obtain  lbl(STa) lbl(STb) | lbl(STc) lbl(T)-lbl(STa U STb U STc)  ########
	    //#    ( ... (((STa),(STb)),STc) ... );
	    //#          ||     |     |    |
	    //#          xv     w    'u'   y
	    v = aop(tr, u); w = apc(tr, u); x = cop(tr, u); y = ccp(tr, u); stdts = new TreeSet<Short>(ts);
	    sta = new short[j=(split=tr.substring(v+1, w).replaceAll("[,\\(\\)]+", " ").trim().split(" ")).length]; for (String s: split) stdts.remove(Short.valueOf(sta[--j]=Short.parseShort(s)));
	    stb = new short[j=(split=tr.substring(++w, u).replaceAll("[,\\(\\)]+", " ").trim().split(" ")).length]; for (String s: split) stdts.remove(Short.valueOf(stb[--j]=Short.parseShort(s)));
	    stc = new short[j=(split=((u+1 != y) ? tr.substring(u+2, y) : tr.substring(++x, --v)).replaceAll("[,\\(\\)]+", " ").trim().split(" ")).length]; for (String s: split) stdts.remove(Short.valueOf(stc[--j]=Short.parseShort(s)));
	    std = new short[j=stdts.size()]; for (Short si: stdts) std[--j] = si.shortValue();
	    //# comparing every quartet ab|cd with the topology induced by dm, where a,b,c,d belongs to STa,STb,STc,STd, respectively  ##########
	    up = 0; //Arrays.sort(sta); Arrays.sort(stb); Arrays.sort(stb); 
	    for (short a: sta) {
		for (short b: stb) { ab = (a>b)?dm[a][b]:dm[b][a];
		    for (short c: stc) { ac = (a>c)?dm[a][c]:dm[c][a]; bc = (b>c)?dm[b][c]:dm[c][b];
			for (short d: std) up += ( (abcd=ab+((c>d)?dm[c][d]:dm[d][c])) < ac+((b>d)?dm[b][d]:dm[d][b])  &&  abcd < bc+((a>d)?dm[a][d]:dm[d][a]) ) ? 1 : 0;
		    }
		}
	    }
	    //# storing re value inside are #####################################################################################################
	    are.add(Double.valueOf(re=up/(dn=((double)sta.length)*((double)stb.length)*((double)stc.length)*((double)std.length))));
	    if ( verbose ) System.out.println((Arrays.toString(sta) + Arrays.toString(stb) + Arrays.toString(stc) + Arrays.toString(std)).replaceAll(" ","") + " Re=" + df.format(re) + " (" + ((long)up) + "/" + ((long)dn) + ")");
	}

	//### writing re values into nwk #####################################################################################################################################################
	u = -1; v = -1; while ( (u=nwk.indexOf(")", ++u)) != -1 ) { if ( nwk.charAt(u+1) == ';' ) break; nwk = nwk.insert(u+1, df.format(are.get(++v).doubleValue())); }
	out = new BufferedWriter(new FileWriter(new File(args[2]))); out.write(nwk.toString()); out.close(); //System.out.println(nwk.toString());
	if ( verbose ) System.out.println("output tree written into " + args[2]);
    }

    //## containing opening parenthesis (cop): returns x from t and i, where i is a closing parenthesis index
    //   t = ( ... ((STa),(STb)) ... );
    //             |          |
    //             x          i
    final static int cop (StringBuilder t, int i) { short p = 0; ++p; int x = i; while ( --x >= 0 ) switch ( t.charAt(x) ) { case ')': ++p; continue; case '(': if ( --p < 0 ) return x; } return -1; }

    //## containing closing parenthesis (ccp): returns x from t and i, where i is a closing parenthesis index
    //   t = ( ... ((STa),(STb)) ... );
    //                  |      |
    //                  i      x
    final static int ccp (StringBuilder t, int i) { short p = 0; int x = i; while ( true ) switch ( t.charAt(++x) ) { case ';': return -1; case '(': ++p; continue; case ')': if ( --p < 0 ) return x; } }

    //## associated opening parenthesis (aop): returns x from t and i, where i is a closing parenthesis index
    //   t = ( ... ((STa),(STb)) ... );
    //             |           |
    //             x           i
    final static int aop (StringBuilder t, int i) { short p = 0; ++p; int x = i; while ( --x >= 0 ) switch ( t.charAt(x) ) { case ')': ++p; continue; case '(': if ( --p == 0 ) return x; } return -1; }

    //## associated partitionning comma (apc): returns x from t and i, where i is a closing parenthesis index
    //   t = ( ... ((STa),(STb)) ... );
    //                   |     |
    //                   x     i
    final static int apc (StringBuilder t, int i) { short p = 0; int x = i; while ( --x >= 0 ) switch ( t.charAt(x) ) { case ')': ++p; continue; case '(': --p; continue; case ',': if ( p == 0 ) return x; } return -1; }
}
